/*
 * FILE: dbgmem.c
 * 
 * ABSTRACT: debug memory functions
 * 
 * provides debug versions of malloc, calloc, realloc and free. All of this use
 * the original libc functions but allocate a little bit more memory to store
 * status information. With this infos it is possible to check pointers for
 * validity.
 * 
 * $Id: dbgmem.c,v 1.3 1999/01/01 15:30:31 rks Exp $
 * 
 * CHANGE LOG:
 * ##  who  ddMmmyy  defect  description
 * --  ---  -------  ------  -------------------------------------------------
 *  1  rks  01Jan99  --      added parameter function name where file & line
 *                           was given
 *  0  rks  29Oct98  --      initial
 */ 
#ifdef USE_DBG_MEM

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "dbgmem.h"

#define DBG_DBG_MEM 0

#define DBG_MEM_FLAG         0xAA
  
#define ulong unsigned long


typedef struct dbg_mcb_t
{
    struct dbg_mcb_t    *mcb_prev;
    struct dbg_mcb_t    *mcb_next;
    char                *mcb_file;
    long                 mcb_line_no;
    char                *mcb_function;
    size_t               mcb_size;
    unsigned char       *mcb_end_flag;
}dbg_mcb_t;

typedef struct dbg_mem_info_t
{
    long        dmc_malloc;
    long        dmc_calloc;
    long        dmc_realloc;
    long        dmc_free;
}dbg_mem_info_t;

static dbg_mem_info_t   dbg_mem_info;
static dbg_mcb_t       *dbg_head = NULL;

static int              initialized = 0;

/*
 * FUNCTION: dbg_mem_init
 * 
 * ABSTRACT: initialise debug memory counters
 * 
 * PARAMETER:
 * NONE
 * 
 * RETURNVALUE:
 * NONE
 */ 
void dbg_mem_init(void)
{
    dbg_mem_info.dmc_malloc = 0;
    dbg_mem_info.dmc_calloc = 0;
    dbg_mem_info.dmc_realloc = 0;
    dbg_mem_info.dmc_free = 0;
    
    initialized = 1;
}

/*
 * FUNCTION: dbg_allocate
 * 
 * ABSTRACT: allocate mem and hook into mcb chain
 * 
 * PARAMETER:
 * caller:      dbg_*alloc function who called dbg_allocate
 * size:        user size of memory needed
 * filename:    file from where function is called
 * line_no:     line from where function is called
 * function:    function from where function is called
 * 
 * RETURNVALUE:
 * NULL:        memory allocation failed
 * !NULL:       ptr to allocated memory
 */ 
void *dbg_allocate(char *caller, size_t size, char *filename, int line_no, 
        char *function)
{
    dbg_mcb_t           *new;
    size_t               real_size = size + sizeof(dbg_mcb_t) + 1;
    char                *ptr;
    
    if(initialized == 0) dbg_mem_init();

    /* do the real allocation */
    new = malloc(real_size);
    if(new == NULL)
    {
        return NULL;
    }
    
    new->mcb_file = filename;
    new->mcb_line_no = line_no;
    new->mcb_function = function;
    new->mcb_size = size;

    /* hook into memory control block chain */
    new->mcb_prev = NULL;
    new->mcb_next = dbg_head;
    dbg_head = new;
    if(new->mcb_next != NULL) new->mcb_next->mcb_prev = new;
    
    /* application pointer */
    ptr = (char *)new + sizeof(dbg_mcb_t);
    
    /* set end of mem flag */
    new->mcb_end_flag = (unsigned char *)(ptr + size);
    *new->mcb_end_flag = DBG_MEM_FLAG;
    
#if DBG_DBG_MEM
    fprintf(stderr, "from %s: 0x%08x = dbg_allocate(%d, %s, %d, %s)\n", caller,
            (unsigned int)ptr, size, filename, line_no, function);
#endif

    return ptr;
}

/*
 * FUNCTION: dbg_malloc
 * 
 * ABSTRACT: debug version of malloc
 * 
 * PARAMETER:
 * size:        user size of memory needed
 * filename:    file from where function is called
 * line_no:     line from where function is called
 * function:    function from where function is called
 * 
 * RETURNVALUE:
 * NULL:        memory allocation failed
 * !NULL:       ptr to allocated memory
 */ 
void *dbg_malloc(size_t size, char *filename, int line_no, char *function)
{
    void                *ptr;
    dbg_mem_info.dmc_malloc++;

    ptr = dbg_allocate("dbg_malloc", size, filename, line_no, function);
    
    return ptr;
}

/*
 * FUNCTION: dbg_calloc
 * 
 * ABSTRACT: debug version of calloc
 * 
 * PARAMETER:
 * count:       number of elements to allocate
 * size:        size of a single element
 * filename:    file from where function is called
 * line_no:     line from where function is called
 * function:    function from where function is called
 * 
 * RETURNVALUE:
 * NULL:        memory allocation failed
 * !NULL:       ptr to allocated memory
 */ 
void *dbg_calloc(size_t count, size_t size, char *filename, int line_no,
        char *function)
{
    void        *ptr;

    dbg_mem_info.dmc_calloc++;

    /* call dbg_malloc */
    ptr = dbg_allocate("dbg_calloc", count * size, filename, line_no,
            function);
    if(ptr == NULL)
    {
        fprintf(stderr, "DBGMEM: failed dbg_calloc(%d, %d, %s, %d, %s)\n",
            count, size, filename, line_no, function);
        return NULL;
    }
    
    memset(ptr, 0, count * size);
    
    return ptr;
}

/*
 * FUNCTION: dbg_realloc
 * 
 * ABSTRACT: debug version of realloc
 * 
 * If p is NULL, behaves like malloc.
 * If size is 0 behaves like free.
 * 
 * PARAMETER:
 * p:           ptr to mem chunk to resize
 * size:        size of a single element
 * filename:    file from where function is called
 * line_no:     line from where function is called
 * function:    function from where function is called
 * 
 * RETURNVALUE:
 * NULL:        memory allocation failed
 * !NULL:       ptr to allocated memory
 */ 
void *dbg_realloc(void *p, size_t size, char *filename, int line_no,
        char *function)
{
    void                *ptr;
    dbg_mcb_t           *mcb;
    int                  real_size = size + sizeof(dbg_mcb_t) + 1;

#if DBG_DBG_MEM
    fprintf(stderr, "dbg_realloc(0x%08x, %d, %s, %d, %s)\n",
            (unsigned int)p, size, filename, line_no, function);
#endif

    if(initialized == 0) dbg_mem_init();
    dbg_mem_info.dmc_realloc++;

    /* is pointer valid ? */
    if((p != NULL) && dbg_mcb_check(p))
    {
        fprintf(stderr, "DbgMem: Realloc: %s:%d (%s): invalid pointer "
                "0x%08lx\n", filename, line_no, function, (unsigned long)p);
        return NULL;
    }
    
    /* if new size is 0 free memory */
    if(size == 0)
    {
        dbg_free(p, filename, line_no, function);
        return NULL;
    }

    /* if p is NULL do normal allocation */
    if(p == NULL)
    {
        return dbg_allocate("dbg_realloc", size, filename, line_no, function);
    }

    mcb = (dbg_mcb_t *)((char *)p - sizeof(dbg_mcb_t));
    
    ptr = realloc(mcb, real_size);
    if(ptr == NULL)
    {
        fprintf(stderr, "DBGMEM: Realloc %s:%d (%s): failed to resize 0x%08lx "
                "to %d byte\n", filename, line_no, function,
                (unsigned long) p, size);
        return NULL;
    }
    mcb = ptr;
    ptr = (char *)mcb + sizeof(dbg_mcb_t);

    /* set end of mem flag */
    mcb->mcb_end_flag = (unsigned char *)(ptr + size);
    *mcb->mcb_end_flag = DBG_MEM_FLAG;

    /* update mcb infos */
    mcb->mcb_size = size;
    if( mcb->mcb_prev != NULL)
    {
        mcb->mcb_prev->mcb_next = mcb;
    }
    else
    {
        dbg_head = mcb;
    }
    if(mcb->mcb_next != NULL)
    {
        mcb->mcb_next->mcb_prev = mcb;
    }

    return ptr;
}


/*
 * FUNCTION: dbg_free
 * 
 * ABSTRACT: debug version of free
 * 
 * handles p == NULL
 * 
 * PARAMETER:
 * p:           ptr to memory to free
 * filename:    file from where function is called
 * line_no:     line from where function is called
 * function:    function from where function is called
 * 
 * RETURNVALUE:
 * NONE
 */ 
void dbg_free(void *p, char *filename, int line_no, char *function)
{
    dbg_mcb_t           *mcb;
    int                  rc;
    
    if(initialized == 0) dbg_mem_init();
    dbg_mem_info.dmc_free++;

    /* if pointer is NULL, there's nothing to do */
    if(p == NULL)
    {
        return;
    }


#if DBG_DBG_MEM
    fprintf(stderr, "dbg_free(0x%08x, %s, %d, %s)\n", (unsigned int)p,
            filename, line_no, function);
#endif
    
    rc = dbg_mcb_check(p);
    if(rc == DBG_MEM_INVALID)
    {
        
        fprintf(stderr, "DBGMEM: Free %s:%d (%s): invalid pointer 0x%08lx\n",
                filename, line_no, function, (unsigned long) p);
        return;
    }
    else if(rc == DBG_MEM_OVERWRITE)
    {
        fprintf(stderr, "DBGMEM: Free %s:%d (%s): detected memory overwrite "
                "for pointer 0x%08lx allocated at %s:%ld : found: 0x%02x "
               "expected: 0x%02x\n",
                filename, line_no, function, (unsigned long) p, mcb->mcb_file,
                mcb->mcb_line_no, *mcb->mcb_end_flag, DBG_MEM_FLAG);
        return;
    }

    /* remove mcb from mcb-chain */
    mcb = (dbg_mcb_t *)((char *)p - sizeof(dbg_mcb_t));

    if(mcb->mcb_prev != NULL)
    {
        mcb->mcb_prev->mcb_next = mcb->mcb_next;
    }
    else
    {
        dbg_head = mcb->mcb_next;
    }
    if(mcb->mcb_next != NULL)
    {
        mcb->mcb_next->mcb_prev = mcb->mcb_prev;
    }
    
    free(mcb);
}

/*
 * FUNCTION: dbg_mem_exit
 * 
 * ABSTRACT: prints all currently allocated memory
 * 
 * PARAMETER:
 * NONE
 * 
 * RETURNVALUE:
 * NONE
 */ 
void dbg_mem_exit(void)
{
    dbg_mcb_t           *mcb = dbg_head;
    
    while(mcb != NULL)
    {
        fprintf(stderr, "DbgMem: memory 0x%08lx allocated at %s:%ld (%s) not "
                "freed %u byte\n",
                (unsigned long)((char *)mcb + sizeof(dbg_mcb_t)), 
                mcb->mcb_file, mcb->mcb_line_no, mcb->mcb_function,
                mcb->mcb_size);
        mcb = mcb->mcb_next;
    }

#if DBG_DBG_MEM
    fprintf(stderr, "Stats:\n");
    fprintf(stderr, "  Malloc:  %6ld\n", dbg_mem_info.dmc_malloc);
    fprintf(stderr, "  Calloc:  %6ld\n", dbg_mem_info.dmc_calloc);
    fprintf(stderr, "  Realloc: %6ld\n", dbg_mem_info.dmc_realloc);
    fprintf(stderr, "  Free:    %6ld\n", dbg_mem_info.dmc_free);
#endif

}

/*
 * FUNCTION: dbg_mcb_check
 * 
 * ABSTRACT: checks if given pointer is valid
 * 
 * PARAMETER:
 * p:           pointer to check
 * 
 * RETURNVALUE:
 * SUCCESS:           pointer is OK
 * DBG_MEM_INVALID:   pointer is INVALID
 * DBG_MEM_OVERWRITE: pointer is OK but detected memory overwrite
 */ 
int dbg_mcb_check(void *p)
{
    dbg_mcb_t           *mcb;
    
    if(p == NULL)
    {
        return 0;
    }

    mcb = (dbg_mcb_t *)((char *)p - sizeof(dbg_mcb_t));

    if((mcb->mcb_prev != NULL) && (mcb->mcb_prev->mcb_next != mcb))
    {
        fprintf(stderr, "PREV test failed\n");
        return DBG_MEM_INVALID;
    }
    if((mcb->mcb_next != NULL) && (mcb->mcb_next->mcb_prev != mcb))
    {
        fprintf(stderr, "NEXT test failed\n");
        return DBG_MEM_INVALID;
    }
    if((mcb->mcb_next == NULL) && (mcb->mcb_prev == NULL) && (dbg_head != mcb))
    {
        fprintf(stderr, "HEAD test failed\n");
        return DBG_MEM_INVALID;
    }
    if(*mcb->mcb_end_flag != DBG_MEM_FLAG)
    {
        return DBG_MEM_OVERWRITE;
    }
    
    return SUCCESS;
}

/*
 * FUNCTION: dbg_ptr_check
 * 
 * ABSTRACT: checks if a ptr points into a allocated memory chunk
 * 
 * PARAMETER:
 * p:           ptr to check
 * 
 * RETURNVALUE:
 * >0:  max bytes to write to this pointer
 * -1:  ptr invalid
 */ 
long dbg_ptr_check(void *p)
{
    dbg_mcb_t           *mcb;
    char                *ptr = p;
    long                 len;
    
    mcb = dbg_head;
    
    while(mcb != NULL)
    {
        if((ptr >= ((char *)mcb + sizeof(dbg_mcb_t))) &&
                (ptr < (char *)mcb->mcb_end_flag))
        {
            len = (char *)mcb->mcb_end_flag - ptr;
            fprintf(stderr, "DBGMEM: PTR OK: in mem allocated: %s:%ld, %ld "
                    "bytes til end of mem block\n", mcb->mcb_file,
                    mcb->mcb_line_no, len);
            return len;
        }
    }

    fprintf(stderr, "DBGMEM: PTR INVALID\n");
    return -1;
}

/*
 * FUNCTION: dbg_mcb_walk
 * 
 * ABSTRACT:
 * 
 * PARAMETER:
 * NONE
 * 
 * RETURNVALUE:
 * SUCCESS:             Memory OK
 * DBG_MEM_INVALID:     Memory corrupted
 * DBG_MEM_OVERWRITE:   Memory seems OK, but mem overwrite detected
 */ 
int dbg_mem_walk(void)
{
    dbg_mcb_t           *mcb = dbg_head;
    int                  rc;
    int                  i = 1;
    int                  mem_over = 0;
    
    while(mcb != NULL)
    {
        rc = dbg_mcb_check((char *)mcb + sizeof(dbg_mcb_t));
        if(rc == DBG_MEM_INVALID)
        {
            
            fprintf(stderr, "DBGMEM: memory corrupted: invalid pointer "
                    "0x%08lx (%d)\n",
                    (unsigned long)((char *)mcb + sizeof(dbg_mcb_t)), i);
            return DBG_MEM_INVALID;
        }
        else if(rc == DBG_MEM_OVERWRITE)
        {
            mem_over = 1;

            fprintf(stderr, "DBGMEM: detected memory overwrite for "
                "pointer 0x%08lx allocated at %s:%ld : found: 0x%02x "
                "expected: 0x%02x\n",
                (unsigned long) ((char *)mcb + sizeof(dbg_mcb_t)),
                mcb->mcb_file, mcb->mcb_line_no, *mcb->mcb_end_flag,
                DBG_MEM_FLAG);
            fprintf(stderr, "DBGMEM: trying to continue...\n");
        }
        mcb = mcb->mcb_next;
        i++;
    }
    
    return mem_over?DBG_MEM_OVERWRITE:SUCCESS;
}

#endif /* USE_DBG_MEM */
/*--------[ END OF FILE dbgmem.c ]-------------------------------------------*/
